#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2010 Christian Dannie Storgaard
#

import gobject, gtk, pango
import vineyard

class SimpleList(gtk.VBox):
    def __init__(self, headers=None, rows=None, types=None):
        gobject.signal_new(
            "toggled",
            self,
            gobject.SIGNAL_ACTION,
            gobject.TYPE_NONE,
            (
                gobject.TYPE_PYOBJECT,
                gobject.TYPE_PYOBJECT,
                gobject.TYPE_PYOBJECT
            ))
        self._fallback_pixbuf = None
        
        # Remember: First cell is index
        if rows is None:
            rows = []
        
        if len(rows):
            self.rows = self.__failsafe_rows(rows)
            self._cell_types = self.__convert_rows_to_types(self.rows)
        elif types is not None:
            self.rows = rows
            self._cell_types = self.__convert_rows_to_types([types])
        else:
            raise(ValueError, "You need to supply either rows or types.")
        
        if headers is None:
            headers = [ '' for i in range(len(self._cell_types)-1) ]
        
        self.headers = headers
        
        gtk.VBox.__init__(self)
        
        self._create_widgets()
        self._create_headers()
        self.fill()
    
    def _create_widgets(self):
        self.scrolled_window = gtk.ScrolledWindow()
        self.scrolled_window.set_border_width(0)
        self.scrolled_window.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.pack_start(self.scrolled_window, expand=True, fill=True)
        
        self.treeview = gtk.TreeView()
        self.scrolled_window.add(self.treeview)
        
        self.model = gtk.ListStore(*self._cell_types)
        self.treeview.set_model(self.model)
        
        self.show_all()
    
    def get(self):
        return list(self.iter())
    
    def iter(self):
        return (
            [ col for col in row ]
            for row in self.model
        )
    
    def set(self, data, index=None, row=None, column=None):
        if index is None and row is None and column is None:
            return self.fill(rows = data)
        
        ## Don't go by row (nr), find by index
        if index is not None:
            for row, _row in enumerate(self.model):
                if _row[0] == index:
                    break
        
        if column is None:
            self.model[row] = data
        else:
            self.model[row][column] = data
    
    def fill(self, rows=None):
        if rows is None:
            rows = self.rows
        
        self.model.clear()
        for row in rows:
            cells = []
            for index, cell in enumerate(row):
                if self._cell_types[index] == gtk.gdk.Pixbuf:
                    pixbuf = vineyard.pixbuf_new_from_string(cell)
                    if pixbuf:
                        cells.append(pixbuf)
                    else:
                        cells.append(self._fallback_pixbuf)
                else:
                    cells.append(cell)
            
            self.model.append(cells)
    
    def set_fallback_pixbuf(self, pixbuf_or_name):
        if type(pixbuf_or_name) in (str, unicode):
            self._fallback_pixbuf = vineyard.pixbuf_new_from_string(pixbuf_or_name)
        else:
            self._fallback_pixbuf = pixbuf_or_name
    
    def _create_headers(self):
        self.columns = []
        for header_title in self.headers:
            self.columns.append(
                gtk.TreeViewColumn(header_title)
            )
            self.columns[-1].set_resizable(True)
            self.treeview.append_column(self.columns[-1])
        
        self.renderers = []
        for column_nr, column_type in enumerate(self._cell_types):
            ## Skip first column, it is index
            if column_nr == 0:
                continue
            
            if column_type == gobject.TYPE_BOOLEAN:
                self.renderers.append(
                    gtk.CellRendererToggle()
                )
                self.columns[column_nr-1].pack_start(self.renderers[-1])
                self.columns[column_nr-1].add_attribute(self.renderers[-1],
                                                      'active', column_nr)
                self.renderers[-1].set_property('activatable', True)
                self.renderers[-1].connect('toggled', self._toggled, column_nr)
                ## Turn of selection if we can toggle stuff, they are confusing
                ## together, if the user does not want this, set it manually
                self.treeview.get_selection().set_mode(gtk.SELECTION_NONE)
            elif column_type == gtk.gdk.Pixbuf:
                self.renderers.append(
                    gtk.CellRendererPixbuf()
                )
                self.columns[column_nr-1].pack_start(self.renderers[-1])
                self.columns[column_nr-1].add_attribute(self.renderers[-1],
                                                      'pixbuf', column_nr)
            else:
                self.renderers.append(
                    gtk.CellRendererText()
                )
                self.renderers[-1].set_property('ellipsize', pango.ELLIPSIZE_END)
                self.columns[column_nr-1].pack_start(self.renderers[-1])
                self.columns[column_nr-1].add_attribute(self.renderers[-1],
                                                      'markup', column_nr)
                self.columns[column_nr-1].set_expand(True)
        if any(self.headers):
            self.treeview.set_headers_visible(True)
        else:
            self.treeview.set_headers_visible(False)
    
    def _toggled(self, renderer, row, column_nr):
        self.model[row][column_nr] = not self.model[row][column_nr]
        self.emit('toggled', row, column_nr, self.model[row][column_nr])
    
    def __failsafe_rows(self, rows):
        new_rows = []
        max_row_length = max(*[ len(row) for row in rows ])
            
        for row in rows:
            # Make sure each row has the same amount of cells
            # adding empty cells to short rows
            new_rows.append(
                row + [ '' for i in range(max_row_length) ][len(row):]
            )
        return new_rows
    
    def __convert_rows_to_types(self, rows):
        cell_types = []
        for cell in rows[0]:
            if cell in (str, unicode) or type(cell) in (str, unicode):
                cell_types.append(gobject.TYPE_STRING)
            elif cell is int or type(cell) is int:
                cell_types.append(gobject.TYPE_INT)
            elif cell is float or type(cell) is float:
                cell_types.append(gobject.TYPE_FLOAT)
            elif cell is bool or type(cell) is bool:
                cell_types.append(gobject.TYPE_BOOLEAN)
            else:
                cell_types.append(gtk.gdk.Pixbuf)
        return cell_types